from typing import Dict


def keyval_to_dict(key, value, prefix="", separator="__", lowercase=False) -> Dict:
    """    Transform key-value into Mapping.    """

    if prefix and not key.startswith(prefix):
        # if prefix is specified, then return nothing for keys without it
        return {}

    # strip key off of prefix
    prefixless_key = key[len(prefix) :]
    # lowercase key if needed
    key = prefixless_key.lower() if lowercase else prefixless_key
    prefix = prefix.lower() if lowercase else prefix

    parts = key.strip(separator).split(separator, maxsplit=1) if separator else key
    if isinstance(parts, list) and len(parts) > 1:
        k, subkey = parts  # unpack split parts for readability
        return {k: keyval_to_dict(subkey, value, prefix="", separator=separator)}
    elif isinstance(parts, list) and len(parts) == 1:
        # Special case for Issue#21
        return {parts[0]: value}
    else:
        return {parts: value}



from ilexconf import Config

import json
import pytest


NESTED = {
    "a1": {"b1": {"c1": 1, "C2": 2, "c3": 3}, "b2": {"c1": "a", "c2": True, "c3": 1.1}},
    "d1": False,
}

MERGABLE = {"a1": {"b1": {"c3": "val"}}}

MERGED = {
    "a1": {
        "b1": {"c1": 1, "C2": 2, "c3": "val"},
        "b2": {"c1": "a", "c2": True, "c3": 1.1},
    },
    "d1": False,
}

WITH_LISTS = {
    "database": {
        "connections": [
            {"host": "1.2.3.4", "port": 5000},
            {"host": "4.5.6.7", "port": 8080},
        ],
        "params": [["kw1", "kw2", 3], {"key": "value of key"}],
    }
}


def test_empty_config():
    cfg = Config()
    assert cfg == {}
    assert "a0" not in cfg


def test_nested_access():
    cfg = Config(NESTED)
    # Plain access
    assert cfg["a1"]["b1"]["c1"] == 1
    assert cfg["a1.b1.c1"] == 1
    assert cfg.a1.b1.c1 == 1
    # Mixed access
    assert cfg["a1"].b1.C2 == 2
    assert cfg.a1["b1"].c3 == 3
    assert cfg.a1.b2["c3"] == 1.1
    # Direct access
    assert cfg.d1 is False
    assert cfg["d1"] is False


def test_simple_merge():
    cfg = Config(Config({"c1": 1, "C2": 2, "c3": 3}), Config({"c3": "val"}))
    assert cfg.as_dict() == {"c1": 1, "C2": 2, "c3": "val"}


def test_nested_merge():
    cfg = Config(NESTED, MERGABLE)
    # "val" has replaced 3 as value for a1.b1.c3
    assert cfg.a1.b1.c3 == "val"
    # other values are still intact
    assert cfg.a1.b2.c3 == 1.1
    assert cfg.a1.b1.c1 == 1


def test_list():
    cfg = Config(WITH_LISTS)
    print(cfg)
    assert cfg.as_dict() == WITH_LISTS


def test_as_dict():
    cfg = Config(NESTED, MERGABLE)
    d = cfg.as_dict()
    assert d == MERGED


def test_setitem():
    cfg = Config(NESTED, MERGABLE)
    cfg["a1"]["b1"]["c1"] = 6
    assert cfg["a1"]["b1"]["c1"] == 6
    assert cfg.a1.b1.c1 == 6

    cfg = Config()
    cfg["f1"] = False
    assert cfg["f1"] is False

    # from .debug import debug
    # debug()
    cfg["f2"]["g3"]["h3"] = 123
    assert cfg["f2.g3.h3"] == 123

    cfg["u1.u4.u6"] = True
    assert cfg["u1"]["u4"]["u6"] is True


def test_setattr():
    cfg = Config(NESTED, MERGABLE)
    cfg.a1.b1.c1 = 8
    assert cfg.a1.b1.c1 == 8


def test_setattr_dict():
    d = {"b": True, "c": {"d": "thename"}}
    cfg = Config()

    # Assign dict to the key. It should atuomatically
    # get transformed into the Config object.
    cfg.a = d
    assert cfg.a.as_dict() == d
    assert cfg.a == Config(d)

    # Notice how the dict got transformed into a Config
    # during assignment so that now we have all the "nice"
    # accessibility features of Config.
    assert cfg.a.c.d == "thename"


def test_setattr_list_with_dict():
    l = [{"b": True, "c": {"d": "thename"}}, "somestring", True]
    cfg = Config()

    # When we assign a list to the key
    # it remains an instance of list, or
    # whatever sequence type it was.
    cfg.a = l
    assert isinstance(cfg.a, list)

    # Internally though any Mapping objects
    # inside that list, no matter how nested,
    # got transformed into the Config objects.
    assert cfg.a == [Config({"b": True, "c": {"d": "thename"}}), "somestring", True]

    # And, again, we get all the nice accessibility
    # shenenigans as a result of such transformation.
    assert cfg.a[0].c.d == "thename"


def test_simple_as_table(settings_json_dict):
    config = Config(settings_json_dict)
    headers, rows = config.as_table()

    assert headers == ["Key", "Value"]
    assert rows == [
        ["database", ""],
        ["  connection", ""],
        ["    host", "localhost"],
        ["    port", "5432"],
    ]


def test_limit_as_table(settings_json_dict):
    config = Config(settings_json_dict)
    headers, rows = config.as_table(limit=3)

    assert rows == [
        ["database", ""],
        ["  connection", ""],
        ["    host", "localhost"],
        ["...", "..."],
    ]


def test_simple_flatten(settings_json_dict):
    config = Config(settings_json_dict)

    flat = config.flatten()
    assert flat == {
        "database.connection.host": "localhost",
        "database.connection.port": 5432,
    }


def test_list_flatten():
    # TODO: implement
    pass


def test_copy(settings_json_dict):
    config = Config(settings_json_dict)

    copy = config.copy()

    # Change value in initial object and check it
    config.database.connection.host = "1.2.3.4"
    assert config.database.connection.host == "1.2.3.4"

    # Make sure value in copied object has not changed
    assert copy.database.connection.host == "localhost"



from cleo import Command, option

from ilexconf.console.options import file_type, scopes


class ConvertCommand(Command):
    name = "convert"
    description = "Convert config from one format to another"

    options = [file_type, *scopes]

    def handle(self):
        print("Nice conversion, my dude. You do that often?")



import debugpy  # pragma: no cover


def debug():  # pragma: no cover
    debugpy.listen(("0.0.0.0", 10001))
    print("⏳ VS Code debugger can now be attached, press F5 in VS Code ⏳", flush=True)
    debugpy.wait_for_client()
    print("🎉 VS Code debugger attached, enjoy debugging 🎉", flush=True)



from cleo import Command, argument, option

from ilexconf.console.options import file_type, scopes


class SetCommand(Command):
    name = "set"

    description = "Set value of the key"
    help = "\n".join(
        [
            "",
            "# Set specific value in the config",
            "set settings.json key value",
            "set settings.json key=value",
            "",
            "# Set environment variable in current shell",
            "set key value",
            "set key=value",
        ]
    )

    arguments = [
        argument(
            name="key",
            description="Config key ('my.key') or key-value pair ('my.key=231')",
        ),
        argument(
            name="value",
            description="[optional] Value of the key when <key> argument is just a key name",
            optional=True,
        ),
    ]

    options = [file_type, *scopes]

    def handle(self):
        print("Wow, you've set it")



from ilexconf import Config

from ilexconf.adapters.env import from_env, to_env
from ilexconf.adapters.ini import from_ini
from ilexconf.adapters.json import from_json, to_json


formats = Config()

formats.json.enabled = True
formats.json.reader = from_json
formats.json.writer = to_json
formats.json.extensions = ["json"]

formats.yaml.enabled = False
formats.yaml.extensions = ["yaml", "yml"]

enabled_formats = [f for f in formats if formats[f].enabled]



from cleo import Command
from pathlib import Path

from ilexconf.config import Config
from ilexconf.adapters import formats, from_env, to_json

from .common_arguments import argument
from .common_options import option


class ListCommand(Command):
    name = "list"

    description = "List configuration file"
    help = "\n".join(
        [
            "",
            "Show contents of the settings.json file as a table",
            "(file format is guessed based on extension):",
            "",
            "    <info>show settings.json</info>",
            "",
            "Show contents of my.conf file but explicitly state",
            "that it contains YAML:",
            "",
            "    <info>show my.conf -t yaml</info>",
            "",
            "Show this help message:",
            "",
            "    <info>show --help</info>",
        ]
    )

    arguments = [argument("path")]
    options = [
        option("type"),
        option("display"),
        option("flatten"),
        option("env"),
        option("prefix"),
        option("separator"),
        option("lowercase"),
        option("uppercase"),
        # option("user"),
        # option("system"),
        # option("tree")
    ]

    aliases = ["show", "print", "describe"]

    def handle(self):
        config = Config()

        filetype = None
        filename = self.argument("path")
        if filename:
            path = Path(filename)

            filetype = self.option("type")
            if filetype and filetype in formats:
                # Use specified filetype if the option was passed
                filetype = formats[filetype]

            elif path.suffix:
                # If file has an extension, try to use to guess the file type
                filetype = self._guess_format_by_suffix(path.suffix)

            if not filetype:
                self.line("<error>Could not determine file type</error>")
                exit()

            try:
                file_config = filetype.reader(filename, read_from_file=True)
                config.merge(file_config)
            except Exception as e:
                self.line(f"<error>Could not parse {filename} file: {e}</error>")
                exit()

        if self.option("env"):
            prefixes = self.option("prefix") or [""]
            for prefix in prefixes:
                params = {"prefix": prefix}
                if self.option("separator"):
                    params["separator"] = self.option("separator")

                try:
                    env_config = from_env(**params)
                    config.merge(env_config)
                except Exception as e:
                    self.line("<error>Could not parse environment variables</error>")
                    exit()

        # Flatten configuration if requested
        if self.option("flatten"):
            config = config.flatten()

        if self.option("uppercase"):
            config = config.upper()

        if self.option("lowercase"):
            config = config.lower()

        # Make sure we have a default writer
        writer = filetype.writer if filetype else to_json

        # Print results using the writer
        output = writer(config)
        print(output)

    def _guess_format_by_suffix(self, suffix: str):
        extension = suffix.strip(".")
        # Build inverted index that maps each extension -> format
        extension_matching = {
            ext: formats[f]
            for f in formats
            for ext in formats[f].extensions
            if formats[f].enabled
        }

        # Guess file format using extension and inverted index
        return (
            extension_matching[extension] if extension in extension_matching else None
        )



import json

from ilexconf.config import Config

from typing import Union, TextIO, Mapping, Any


def from_json(data: Union[str, TextIO], read_from_file: bool = True):
    """
    Read data from JSON file or JSON string.
    """
    if read_from_file:
        if isinstance(data, str):
            json_dict = json.load(open(data, "rt"))
        else:
            json_dict = json.load(data)
    else:
        json_dict = json.loads(str(data))

    config = Config(json_dict)
    return config


def to_json(data: Mapping[Any, Any], path: str = None, indent: int = 2):
    """
    Convert data to JSON and optionally write to file.
    """
    if isinstance(data, Config):
        data = data.as_dict()

    # Convert data to json
    json_data = json.dumps(data, indent=indent)

    if path:
        with open(path, "w") as f:
            f.write(json_data)

    return json_data



from cleo import Command, argument, option

from ilexconf.console.options import file_type, scopes


class GetCommand(Command):
    name = "get"
    description = "Get value from the configuration"

    arguments = [
        argument(
            name="key",
            description="Key to retrieve",
            optional=False,
        )
    ]

    options = [file_type, *scopes]

    def handle(self):
        print("Here comes value")



from ilexconf.helpers import keyval_to_dict

from typing import Any, Dict, Mapping, List, Sequence


class Config(dict):
    """
    Config is a dictionary of other configs forming hierarchical structure.
    """

    def __init__(self, *mappings: Mapping[Any, Any], **kwargs: Dict):
        """
        Constructor.
        """

        super().__init__()

        # Merge in values of mappings
        self.merge(*mappings, **kwargs)

    def __getitem__(self, item):
        if isinstance(item, str) and "." in item:
            key, subkey = item.split(".", maxsplit=1)
            return self._dd_getitem(key).__getitem__(subkey)
        else:
            return self._dd_getitem(item)

    def __getattr__(self, attr):
        return self._dd_getitem(attr)

    def __setitem__(self, item, value):
        value = self._parse(value)
        if isinstance(item, str) and "." in item:
            key, subkey = item.split(".", maxsplit=1)
            self._dd_getitem(key).__setitem__(subkey, value)
        else:
            dict.__setitem__(self, item, value)

    def __setattr__(self, attr, value):
        value = self._parse(value)
        dict.__setitem__(self, attr, value)

    def __repr__(self):
        return f"Config{dict.__repr__(self)}"

    def merge(self, *mappings: Mapping[Any, Any], **kwargs) -> None:
        """
        Merge values of mappings with current config recursively.
        """

        # For every key of that mapping
        for mapping in mappings:
            for key, value in mapping.items():

                parsed = self._parse(value)
                self.update(
                    {
                        key: Config(self[key], parsed)
                        if key in self
                        and isinstance(self[key], Config)
                        and isinstance(parsed, Config)
                        else parsed
                    }
                )

        # Merge in values of keyword arguments
        for k, v in kwargs.items():
            keyval_dict = keyval_to_dict(k, v)
            self.merge(keyval_dict)

    def flatten(self, prefix="", separator="."):
        """Flatten current config so that there is no hierarchy."""

        d = dict()
        p = f"{prefix}{separator}" if prefix else ""
        for key in self.keys():
            if isinstance(self[key], Config):
                flattened = self[key].flatten(prefix=f"{p}{key}", separator=separator)
                d.update(flattened)
            else:
                d[f"{p}{key}"] = self[key]
        return Config(d)

    def lower(self):
        """Lowercase all string keys of the configuration"""

        return Config(self.as_dict(lowercase=True))

    def upper(self):
        """Uppercase all string keys of the configuration"""

        return Config(self.as_dict(uppercase=True))

    def copy(self):
        """Return deep copy of the Config object."""

        return Config(self.as_dict())

    def as_dict(self, lowercase: bool = False, uppercase: bool = False):
        """Convert configuration to dict object.

        `lowercase` takes precedence over `uppercase`.
        """

        d = dict()
        for key in self.keys():
            assignment_key = (
                key.lower()
                if lowercase and isinstance(key, str)
                else key.upper()
                if uppercase and isinstance(key, str)
                else key
            )
            if isinstance(self[key], Config):
                d[assignment_key] = self[key].as_dict(
                    lowercase=lowercase, uppercase=uppercase
                )
            else:
                d[assignment_key] = self[key]
        return d

    def as_table(
        self,
        headers: List[str] = ["Key", "Value"],
        indentation: str = "  ",
        limit: int = 0,
    ):
        """Transform configuration into table structure.

        Returns tuple of headers and rows.
        Returned structure is suitable for printing by Cleo library.
        """

        def processor(c: Dict, rows: List, level: int):
            # Increment level so that all child rows are indented
            level += 1
            for k in sorted(c.keys()):
                key = k
                table_key = f"{indentation * level}{key}"

                # Try to interpret the field as a dictionary
                try:
                    # Can't do isinstance(c, collections.Mapping) because
                    # config that needs this method is not a Mapping subclass
                    is_dict = hasattr(c[k], "keys")
                except KeyError:  # pragma: no cover
                    is_dict = False

                if limit and len(rows) >= limit:
                    break

                if is_dict:
                    # Add category row
                    rows.append([table_key, ""])
                    # Recursive call
                    processor(c[k], rows, level)

                else:
                    # Transform multiline strings into singleline
                    singleline = " ".join(
                        str(c[k]).replace("\n", " ").replace("\t", " ").split()
                    )

                    # Clip long string by character limit to fit onto the screen
                    clipped = (
                        singleline if len(singleline) < 80 else f"{singleline[:76]} ..."
                    )
                    rows.append([table_key, clipped])

        # Populate rows
        rows = []
        processor(self, rows, -1)
        if limit and len(rows) >= limit:
            rows.append(["...", "..."])

        return (headers, rows)

    def _parse(self, value: Any):
        # If value is another Mapping: dict, Config, etc.
        if isinstance(value, Mapping):
            return Config(value)

        # If value is a Sequence but not str, bytes, or bytearray
        elif isinstance(value, Sequence) and not isinstance(
            value, (str, bytes, bytearray)
        ):
            l = list()
            for i in value:
                l.append(self._parse(i))
            # Return sequence with the same type
            t = type(value)
            return t(l)

        # If value is anything else
        else:
            return value

    def _dd_getitem(self, item):
        """Implements defaultdict feature

        DefaultDict getitem method
        """
        if item not in self:
            dict.__setitem__(self, item, Config())
        return dict.__getitem__(self, item)



from cleo import argument as cleo_argument

from ilexconf import Config


common_args = Config()


def _add_arg(arg):
    common_args[arg.name].arg = arg


def argument(name):
    return common_args[name].arg


path = cleo_argument(
    name="path",
    description="[optional] Path to the configuration file.",
    optional=True,
)
_add_arg(path)



import os
import pytest

from ilexconf import Config, from_json, to_json

# from ilexconf.tests.debug import debug


@pytest.fixture(scope="module")
def resulting_dict():
    return {
        "database": {
            "connection": {
                "host": "test.local",
                "port": 8080,
                "user": "root",
                "password": "different secret",
            }
        }
    }


def test_quick_start(
    settings_json_dict, settings_json_file_path, resulting_dict, tmp_path
):
    # os.putenv("AWS_DEFAULT_REGION", "us-east-1")
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"

    # [create]
    from ilexconf import Config, from_json, from_env, to_json

    # Empty config
    config = Config()
    assert config.as_dict() == {}

    # Create config from json and merge it into our initial config
    # Let settings_json_file_path = "settings.json" where inside the file we have
    # { "database": { "connection": { "host": "localhost", "port": 5432 } } }
    config.merge(from_json(settings_json_file_path))
    assert config.as_dict() == {
        "database": {"connection": {"host": "localhost", "port": 5432}}
    }

    # Merge dict into config
    config.merge({"database": {"connection": {"host": "test.local"}}})
    assert config.as_dict() == {
        "database": {"connection": {"host": "test.local", "port": 5432}}
    }

    # Merge environment variables into config
    config.merge(from_env(prefix="AWS_", separator="__", lowercase=True))
    assert config.as_dict() == {
        "database": {"connection": {"host": "test.local", "port": 5432}},
        "default_region": "us-east-1",
    }

    # Merge keyword arguments
    config.merge(my__keyword__argument=True)
    assert config.as_dict() == {
        "database": {"connection": {"host": "test.local", "port": 5432}},
        "default_region": "us-east-1",
        "my": {"keyword": {"argument": True}},
    }

    # Clear values, just like with dict
    config.clear()
    assert config.as_dict() == {}

    # Or, better yet, do this all in one step, since Config() constructor
    # accepts any number of mapping objects and keyword arguments as
    # initialization parameters. However, order of parameters matters.
    # Last mappings are merged on top of others. And keywords override even that.
    config = Config(
        from_json(settings_json_file_path),
        {"database": {"connection": {"host": "test.local"}}},
        database__connection__port=4000,
    )
    assert config.as_dict() == {
        "database": {"connection": {"host": "test.local", "port": 4000}}
    }
    # [create]

    # from ilexconf import (
    #     from_json,
    #     # from_yaml,
    #     # from_toml,
    #     from_ini,
    #     # from_python,
    #     # from_dotenv,
    #     from_env,
    # )
    # [read]
    cfg1 = from_json(settings_json_file_path)
    assert cfg1.as_dict() == {
        "database": {"connection": {"host": "localhost", "port": 5432}}
    }
    # [read]

    # cfg2 = Config(
    #     from_yaml("settings.yaml"),
    #     from_toml("settings.toml")
    # )

    # cfg3 = Config(
    #     from_ini("settings.ini"),
    #     from_python("settings.py"),
    #     from_dotenv(".env"),
    #     from_env()
    # )

    # [access]
    # Classic way
    assert config["database"]["connection"]["host"] == "test.local"

    # Dotted key notation
    assert config["database.connection.host"] == "test.local"

    # Via attributes
    assert config.database.connection.host == "test.local"

    # Any combination of the above
    assert config["database"].connection.host == "test.local"
    assert config.database["connection.host"] == "test.local"
    assert config.database["connection"].host == "test.local"
    assert config.database.connection["host"] == "test.local"
    # [access]

    # [upsert]
    # Change value that already exists in the dictionary
    # just like you would do with simple dict
    config["database"]["connection"]["port"] = 8080
    assert config["database"]["connection"]["port"] == 8080

    # Create new value using 'dotted notation'. Notice that
    # 'user' field did not exist before.
    config["database.connection.user"] = "root"
    assert config["database.connection.user"] == "root"

    # Create new value using. 'password' field did not exist
    # before we assigned a value to it and was created automatically.
    config.database.connection.password = "secret stuff"
    assert config.database.connection.password == "secret stuff"
    # [upsert]

    # [merge]
    # Config correctly merges nested values. Notice how it overrides
    # the value of the 'password' key in the nested 'connection' config
    # from 'secret stuff' to 'different secret'
    config.database.connection.merge({"password": "different secret"})
    assert config.database.connection.password == "different secret"
    # [merge]

    # [smart-merge]
    merged = Config({"a1": {"c1": 1, "c2": 2, "c3": 3}}, {"a1": {"c3": "other"}})

    # Instead of overriding the value of the "a1" key completely, `merge` method
    # will recursively look inside and merge nested values.
    assert merged.as_dict() == {"a1": {"c1": 1, "c2": 2, "c3": "other"}}
    # [smart-merge]

    # [as-dict]
    assert config.as_dict() == {
        "database": {
            "connection": {
                "host": "test.local",
                "port": 8080,
                "user": "root",
                "password": "different secret",
            }
        }
    }
    # [as-dict]

    # [write]
    # Temporary path
    p = tmp_path / "settings.json"
    # Save config
    to_json(config, str(p))
    # Verify written file is correct
    assert from_json(str(p)).as_dict() == {
        "database": {
            "connection": {
                "host": "test.local",
                "port": 8080,
                "user": "root",
                "password": "different secret",
            }
        }
    }
    # [write]

    # [subclass]
    class MyConfig(Config):
        def __init__(self, do_stuff=False):
            # Initialize your custom config using json settings file
            super().__init__(self, from_json(settings_json_file_path))

            # Add some custom value depending on some logic
            if do_stuff:
                # Here, we create new nested key that did not exist
                # before and assign a value to it.
                self.my.custom.key = "Yes, do stuff"

            # Merge one more mapping on top
            self.merge({"Horizon": "Up"})

    # [subclass]

    # [test-subclass]
    # Now you can use your custom defined Config. Given the `setting.json` file that
    # contains { "database": { "connection": { "host": "localhost", "port": 5432 } } }
    # MyConfig will have the following values:

    config = MyConfig(do_stuff=True)
    print(config.as_dict())
    assert config.as_dict() == {
        "database": {
            "connection": {
                "host": "localhost",
                "port": 5432,
            },
        },
        "Horizon": "Up",
        "my": {"custom": {"key": "Yes, do stuff"}},
    }
    # [test-subclass]



from cleo import Command

from .list import ListCommand

# from .get import GetCommand
# from .set import SetCommand
# from .convert import ConvertCommand


class ConfigCommand(Command):
    name = "config"
    description = "Configuration command"

    commands = [
        ListCommand(),
        # GetCommand(),
        # SetCommand(),
        # ConvertCommand(),
    ]

    def handle(self):
        return self.call("help", self._config.name)



import os
import pytest


@pytest.fixture(scope="session", autouse=True)
def fixture_dir():
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), "files")


@pytest.fixture(scope="session", autouse=True)
def settings_json_dict():
    return {"database": {"connection": {"host": "localhost", "port": 5432}}}


@pytest.fixture(scope="session", autouse=True)
def settings_json_file_path(fixture_dir):
    return os.path.join(fixture_dir, "settings.json")



import os

from ilexconf.config import Config
from ilexconf.helpers import keyval_to_dict

from typing import Mapping, Any, Union


def from_env(prefix="", separator="__", lowercase=False):
    """Read config from environment variables.

    :param prefix: Prefix to environment variables, defaults to ""
    :type prefix: str, optional
    :param separator: Delimiter by which variable is separated and turned into a nested object, defaults to "__"
    :type separator: str, optional
    :param lowercase: Convert names of variable to lowercase, defaults to False
    :type lowercase: bool, optional
    :return: Config object
    :rtype: Config
    """
    prefix = "" if prefix is None else prefix
    config = Config()

    for k, v in os.environ.items():
        # Convert current key-value pair to Mapping
        d = keyval_to_dict(
            k, v, prefix=prefix, separator=separator, lowercase=lowercase
        )
        # Merge this Mapping into config
        config.merge(d)

    return config


def to_env(
    data: Mapping[Any, Any], prefix="", separator="__", uppercase=False, lowercase=False
):
    """Export mapping to environment variables.

    This will set the environment variable both for current python
    context as well as for child processes invoed via os.system(),
    popen(), fork(), and execv().

    :param data: Mapping object to export.
    :type data: Mapping[Any, Any]
    """

    config = Config(data)

    # Convert entire hierarchy to flat keys with values.
    # { "database": { "connection": { "host": "1.2.3.4" }}} will turn into
    # { "my_database__connection__host": "1.2.3.4" } with prefix == "my_"
    # and separator == "__".
    flat = config.flatten()
    prefix = prefix.upper() if uppercase else prefix.lower() if lowercase else prefix

    for k, v in flat.items():
        adjusted_k = k.upper() if uppercase else k.lower() if lowercase else k
        key = prefix + adjusted_k.replace(".", separator)

        # Set environment variable for current context
        os.environ[key] = v
        os.putenv(key, v)



try:
    import cleo
except ImportError:
    cleo = None

if cleo:
    from .commands.list import ListCommand

    # from .commands.get import GetCommand
    # from .commands.set import SetCommand
    # from .commands.convert import ConvertCommand

    class Application:
        def __init__(self):
            self._app = cleo.Application(name="ilexconf", version="0.6")
            self._app.add(ListCommand())

        def run(self):
            # import debugpy
            # debugpy.listen(10001)
            # debugpy.wait_for_client()
            self._app.run()


else:
    import argparse

    class Application:
        def __init__(self):
            self._parser = argparse.ArgumentParser()

        def run(self):
            self._parser.parse_args()


if __name__ == "__main__":
    Application().run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ilexconf',
 'ilexconf.adapters',
 'ilexconf.console',
 'ilexconf.console.commands',
 'ilexconf.tests']

package_data = \
{'': ['*'], 'ilexconf.tests': ['files/*']}

extras_require = \
{'console': ['cleo>=0.8.1,<0.9.0'], 'yaml': ['ruamel.yaml>=0.16.12,<0.17.0']}

entry_points = \
{'console_scripts': ['ilexconf = ilexconf.console:main']}

setup_kwargs = {
    'name': 'ilexconf',
    'version': '0.6.31',
    'description': 'Configuration library for Python 🔧 Load from multiple sources',
    'long_description': '<a href="https://ilexconf.com"><img alt="ilexconf" src="https://raw.githubusercontent.com/ilexconf/ilexconf/master/docs/_static/github-logo.png"></a>\n\n<h2 align="center">Configuration Library 🔧 for Python</h2>\n\n<p align="center">\n<a href="https://travis-ci.org/ilexconf/ilexconf"><img alt="Build status of package" src="https://img.shields.io/travis/ilexconf/ilexconf?logo=travis"></a>\n<a href="https://github.com/ilexconf/ilexconf/actions?query=workflow%3Adocs"><img alt="Build status of GitHub pages docs" src="https://img.shields.io/github/workflow/status/ilexconf/ilexconf/docs?label=docs&logo=github"></a>\n<a href="https://ilexconf.readthedocs.io/"><img alt="Build status of Read the Docs" src="https://img.shields.io/readthedocs/ilexconf?label=readthedocs&logo=read-the-docs"></a>\n<a href="https://codecov.io/gh/ilexconf/ilexconf"><img alt="Code coverage report" src="https://img.shields.io/codecov/c/github/ilexconf/ilexconf?logo=codecov"></a>\n<a href="https://pypi.org/project/ilexconf/"><img alt="PyPI" src="https://img.shields.io/pypi/v/ilexconf?logo=pypi&color=blue"></a>\n</p>\n\n`ilexconf` is a Python library to load and merge configs from multiple sources, access & change the values, and write them back, if needed. It has no dependencies by default but provides additional functions, relying on popular libraries to parse `yaml`, `toml`, provide `CLI` app, etc.\n\n📝 Full documentation at [ilexconf.com](https://ilexconf.com)\n\n## Table of contents \n\n* <a href="#quick_start">🚀 Quick Start</a>\n  * <a href="#quick_start_install">Installation</a>\n  * <a href="#quick_start_create">Create `Config` object</a>\n  * <a href="#quick_start_read">Read using `from_` functions</a>\n  * <a href="#quick_start_access">Access values</a>\n  * <a href="#quick_start_change_create">Change & create values</a>\n  * <a href="#quick_start_merge">`merge` another `Mapping` into config</a>\n  * <a href="#quick_start_as_dict">Convert to simple `dict` using `as_dict`</a>\n  * <a href="#quick_start_write">Save to file using `to_` functions</a>\n  * <a href="#quick_start_subclass">Subclass `Config` to customize</a>\n* <a href="#internals">⚙️ Internals – How it Works</a>\n  * <a href="#internals_implementation">Subclassing `dict` and overriding methods</a>\n  * <a href="#internals_documentation">Building and hosting docs</a>\n\n<a id="quick_start"></a>\n## 🚀 Quick Start\n\n<a id="quick_start_install"></a>\n### Install\n\n```shell\n$ pip install ilexconf\n```\n\n<a id="quick_start_create"></a>\n### Populate Config with values\n\nConfig object is initialized using arbitrary number of Mapping objects and keyword arguments. It can even be empty. \n\n```python\nfrom ilexconf import Config, from_json\n\n# All of these are valid methods to initialize a config\nconfig = Config()\nconfig = Config({ "database": { "connection": { "host": "test.local" } } })\nconfig = Config(database__connection__port=4000)\nconfig = from_json("settings.json")\nconfig = from_env()\n\n# Or, you can combine them\nconfig = Config(\n    # Take the basic settings from JSON file\n    from_json("settings.json"),\n\n    # Merge the dictionary into that\n    { "database": { "connection": { "host": "test.local" } } },\n\n    # Merge the keyword arguments on top\n    database__connection__port=4000\n)\n```\n\nWhen we initialize config all the values are merged. Arguments are merged in order. Every next argument is merged on top of the previous mapping values. And keyword arguments override even that. _For more details read about <a href="#quick_start_merge">merging</a> strategy below_.\n\nFor a settings file `settings.json` with the following content ...\n\n```json\n{\n    "database": {\n        "connection": {\n            "host": "localhost",\n            "port": 5432\n        }\n    }\n}\n```\n\nThe code above will produce a merged `config` with merged values:\n\n```json\n{\n    "database": {\n        "connection": {\n            "host": "test.local",\n            "port": 4000\n        }\n    }\n}\n```\n\n<a id="quick_start_read"></a>\n### Read from files & environment variables\n\nFiles like `.json`, `.yaml`, `.toml`, `.ini`, `.env`, `.py` as well as environment variables can all be read & loaded using a set of `from_` functions.\n\n```python\nfrom ilexconf import (\n    from_json,      # from JSON file or string\n    from_yaml,      # from YAML file or string\n    from_toml,      # from TOML file or string\n    from_ini,       # from INI file or string\n    from_python,    # from .py module\n    from_dotenv,    # from .env file\n    from_env        # from environment variables\n)\n\ncfg1 = from_json("settings.json")\n\ncfg2 = Config(\n    from_yaml("settings.yaml"),\n    from_toml("settings.toml")\n)\n\ncfg3 = Config(\n    from_ini("settings.ini"),\n    from_python("settings.py"),\n    from_dotenv(".env"),\n    from_env()\n)\n```\n\n<a id="quick_start_access"></a>\n### Access values however you like\n\nYou can access any key in the hierarchical structure using classical Python dict notation, dotted keys, attributes, or any combination of this methods.\n\n```python\n# Classic way\nassert config["database"]["conection"]["host"] == "test.local"\n\n# Dotted key\nassert config["database.connection.host"] == "test.local"\n\n# Attributes\nassert config.database.connection.host == "test.local"\n\n# Any combination of the above\nassert config["database"].connection.host == "test.local"\nassert config.database["connection.host"] == "test.local"\nassert config.database["connection"].host == "test.local"\nassert config.database.connection["host"] == "test.local"\n```\n\n<a id="quick_start_change_create"></a>\n### Change existing values and create new ones\n\nSimilarly, you can set values of any key (_even if it doesn\'t exist in the Config_) using all of the ways above.\n\n**Notice**, _contrary to what you would expect from the Python dictionaries, setting nested keys that do not exist is **allowed**_.\n\n```python\n# Classic way\nconfig["database"]["connection"]["port"] = 8080\nassert config["database"]["connection"]["port"] == 8080\n\n# Dotted key (that does not exist yet)\nconfig["database.connection.user"] = "root"\nassert config["database.connection.user"] == "root"\n\n# Attributes (also does not exist yet)\nconfig.database.connection.password = "secret stuff"\nassert config.database.connection.password == "secret stuff"\n```\n\n<a id="quick_start_merge"></a>\n### Merge with another Mapping object\n\nIf you just assign a value to any key, you override any previous value of that key.\n\nIn order to merge assigned value with an existing one, use `merge` method.\n\n```python\nconfig.database.connection.merge({ "password": "different secret" })\nassert config.database.connection.password == "different secret"\n```\n\n`merge` respects the contents of each value. For example, merging two dictionaries with the same key would not override that key completely. Instead, it will recursively look into each key and try to merge the contents. Take this example:\n\n```python\nconfig = Config(\n    { "a1": { "c1": 1, "c2": 2, "c3": 3 } },\n    { "a1": { "c3": "other" } }\n)\n\n# Instead of overriding the value of the "a1" key completely, `merge` method\n# will recursively look inside and merge nested values.\nassert config.as_dict() == { "a1": { "c1": 1, "c2": 2, "c3": 3 } }\n```\n\n<a id="quick_start_as_dict"></a>\n### Represent as dictionary\n\nFor any purposes you might find fit you can convert entire structure of the Config object into dictionary, which will be essentially returned to you as a deep copy of the object.\n\n```python\nassert config.as_dict() == {\n    "database": {\n        "connection": {\n            "host": "test.local",\n            "port": 8080,\n            "user": "root",\n            "password": "different secret"\n        }\n    }\n}\n```\n\n<a id="quick_start_write"></a>\n### Write to file\n\nYou can serialize the file as JSON or other types any time using the `to_` functions.\n\n```python\n# Write updated config back as JSON file\nfrom ilexconf import to_json\n\nto_json(config, "settings.json")\n```\n\n**WARNING**: _This might throw a serialization error if any of the values contained in the Config are custom objects that cannot be converted to `str`. Also, obviously, you might not be able to correctly parse an object back, if it\'s saved to JSON as `MyObject(<function MyObject.__init__.<locals>.<lambda> at 0x108927af0>, {})` or something._\n\n<a id="quick_start_subclass"></a>\n### Subclass\n\nSubclassing `Config` class is very convenient for implementation of your own config classes with custom logic.\n\nConsider this example:\n\n```python\nimport ilexconf\n\nclass Config(ilexconf.Config):\n    """\n    Your custom Configuration class\n    """\n\n    def __init__(self, do_stuff=False):\n        # Initialize your custom config with JSON by default\n        super().__init__(self, ilexconf.from_json("setting.json"))\n\n        # Add some custom value depending on some logic\n        if do_stuff:\n            self.my.custom.key = "Yes, do stuff"\n\n        self.merge({\n            "Horizon": "Up"\n        })\n\n# Now you can use your custom Configuration everywhere\nconfig = Config(do_stuff=True)\nassert config.my.custom.key == "Yes, do stuff"\nassert config.Horizon == "Up"\n```\n\n<a id="internals"></a>\n## ⚙️ Internals\n\n<a id="internals_implementation"></a>\n### Implementation\n\nUnder the hood `ilexconf` is implemented as a `defaultdict` where every key with Mapping value is represented as another `Config` object. This creates a hierarchy of `Config` objects.\n\n`__getitem__`, `__setitem__`, `__getattr__`, and `__setattr__` methods are overloaded with custom logic to support convenient get/set approach presented by the library.\n\n<a id="internals_documentation"></a>\n### Documentation\n\nFull documentation is available at **[ilexconf.com](https://ilexconf.com)** and on ilexconf.readthedocs.io.\n\n* Documentation is written using `reStructuredText` and uses real code snippets from the unit tests and source code.\n* Documentation is built using [Sphinx](https://www.sphinx-doc.org/) using [sphinx-material](https://github.com/bashtage/sphinx-material) theme.\n* Documentation is hosted on Github Pages at [ilexconf.com](https://ilexconf.com) and on "Read the Docs" at [ilexconf.readthedocs.io](https://ilexconf.readthedocs.io/).\n* For Github Pages documentation is built using Github Actions.\n* Read the Docs builds their version automatically based on the `.readthedocs.yml` config in the project root directory.\n\n## Contributing\n\nContributions are welcome!\n\n## Kudos\n\n`ilexconf` heavily borrows from `python-configuration` library and is inspired by it.\n\n## License\n\nMIT',
    'author': 'vduseev',
    'author_email': 'vagiz@duseev.com',
    'maintainer': 'vduseev',
    'maintainer_email': 'vagiz@duseev.com',
    'url': 'https://github.com/ilexconf/ilexconf',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from cleo import option as cleo_option

from ilexconf import Config
from ilexconf.adapters import enabled_formats


common_opts = Config()


def _add_option(o, choices=None):
    common_opts[o.long_name].opt = o
    if choices:
        common_opts[o.long_name].opt._description += f"Choices: {', '.join(choices)}."
        common_opts[o.long_name].choices = choices


def option(name):
    if name not in common_opts:
        raise Exception(f"Option {name} was not defined")
    return common_opts[name].opt


filetype = cleo_option(
    long_name="type",
    short_name="t",
    description="File type. ",
    flag=False,
    value_required=True,
)
_add_option(filetype, choices=enabled_formats)

source_filetype = cleo_option(
    long_name="source-type",
    short_name="s",
    description="Source file type. ",
    flag=False,
    value_required=True,
)
_add_option(source_filetype, choices=enabled_formats)

destination_filetype = cleo_option(
    long_name="dest-type",
    short_name="d",
    description="Destination file type. ",
    flag=False,
    value_required=True,
)
_add_option(destination_filetype, choices=enabled_formats)

display = cleo_option(
    long_name="display",
    short_name="d",
    description="Display format of the config. ",
    flag=False,
    value_required=True,
)
_add_option(display, choices=["table"] + enabled_formats)

flatten = cleo_option(
    long_name="flatten",
    description="Flatten config before output",
    flag=True,
    value_required=False,
)
_add_option(flatten)

lowercase = cleo_option(
    long_name="lowercase",
    description="Lowercase config keys",
    flag=True,
    value_required=False,
)
_add_option(lowercase)

uppercase = cleo_option(
    long_name="uppercase",
    description="Uppercase config keys",
    flag=True,
    value_required=False,
)
_add_option(uppercase)

prefix = cleo_option(
    long_name="prefix",
    description="Prefix of the environment variables (default: '')",
    flag=False,
    value_required=True,
    multiple=True,
)
_add_option(prefix)

separator = cleo_option(
    long_name="separator",
    description="Environment variable hierarchical delimiter (default: '__')",
    flag=False,
    value_required=True,
)
_add_option(separator)

env = cleo_option(
    long_name="env",
    short_name="e",
    description="Consider environment variables",
    flag=True,
    value_required=False,
)
_add_option(env)

user = cleo_option(
    long_name="user",
    description="Search for config file in user's scope",
    flag=True,
    value_required=False,
)
_add_option(user)

system = cleo_option(
    long_name="system",
    description="Search for config file in system's scope",
    flag=True,
    value_required=False,
)
_add_option(system)

tree = cleo_option(
    long_name="tree",
    description="Search for config file in the directory hierarchy up to the root",
    flag=True,
    value_required=False,
)
_add_option(tree)



import configparser

from typing import Union, TextIO


def from_ini(
    data: Union[str, TextIO], read_from_file: bool = False, lowercase_keys: bool = False
):
    """
    Read data from INI file or INI string.
    """
    if read_from_file:
        if isinstance(data, str):
            data = open(data, "rt").read()
        else:
            data = data.read()
    data = str(data)
    cfg = configparser.RawConfigParser()
    cfg.read_string(data)
    result = {
        section + "." + k: v
        for section, values in cfg.items()
        for k, v in values.items()
    }



import os
import pytest

from ilexconf import Config, from_env, to_env

# from ilexconf.tests.debug import debug


def test_from_env_basic():
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"

    # Prefix = AWS_
    # Separator = __
    # Lowercase = True
    config = from_env(prefix="AWS_", separator="__", lowercase=True)
    assert config.as_dict() == {"default_region": "us-east-1"}

    # Prefix = "" (empty)
    # Separator = ""
    # Lowercasee = False (default)
    # TODO: Tell about environment variables dashes converted to underscores in docs
    config = from_env()
    assert config.AWS_DEFAULT_REGION == "us-east-1"


def test_from_env_no_prefix():
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"

    # debug()
    config = from_env(separator="_", lowercase=True)
    with pytest.raises(AssertionError):
        # Should raise error, because correct key is config.aws.deault.region
        assert config.default.region == "us-east-1"
    assert config.aws.default.region == "us-east-1"

    config = from_env(lowercase=True)
    assert config.aws_default_region == "us-east-1"

    config = from_env(prefix=None, lowercase=True)
    assert config.aws_default_region == "us-east-1"

    config = from_env()
    with pytest.raises(AssertionError):
        # Should raise error, because correct key is uppercase
        assert config.aws_default_region == "us-east-1"
    assert config.AWS_DEFAULT_REGION == "us-east-1"


def test_from_env_no_separator():
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"

    config = from_env(prefix="_", lowercase=True)
    with pytest.raises(AssertionError):
        assert config.aws_default_region == "whatever"
    assert config.aws_default_region == Config()

    # debug()
    config = from_env(separator="", lowercase=True)
    with pytest.raises(AssertionError):
        assert config.AWS_DEFAULT_REGION == "us-east-1"
    assert config.aws_default_region == "us-east-1"

    config = from_env(separator=None)
    with pytest.raises(AssertionError):
        assert config.aws_default_region == "us-east-1"
    assert config.AWS_DEFAULT_REGION == "us-east-1"

    config = from_env()
    with pytest.raises(AssertionError):
        assert config.aws_default_region == "us-east-1"
    assert config.AWS_DEFAULT_REGION == "us-east-1"


def test_from_env_one_dash_separator():
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"

    config = from_env(separator="_")
    with pytest.raises(AssertionError):
        assert config.aws.default.region == "us-east-1"
    assert config.AWS.DEFAULT.REGION == "us-east-1"

    config = from_env(separator="_", lowercase=True)
    with pytest.raises(AssertionError):
        assert config.AWS.DEFAULT.REGION == "us-east-1"
    assert config.aws.default.region == "us-east-1"


def test_to_env_current_context():
    # TODO: Implement
    pass

    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"

    config = from_env(lowercase=True)
    config.aws_default_region = "us-east-2"
    to_env(config, prefix="", uppercase=True)
    assert os.environ["AWS_DEFAULT_REGION"] == "us-east-2"

    to_env(config, prefix="MY", lowercase=False)
    assert os.environ["MYaws_default_region"] == "us-east-2"

    to_env(config, prefix="mY_")
    assert os.environ["mY_aws_default_region"] == "us-east-2"

    to_env(config, prefix="MY_", uppercase=True)
    with pytest.raises(KeyError):
        # Should not exist because MY_ is prefix
        assert os.environ["MY__AWS_DEFAULT_REGION"] == "us-east-2"
    assert os.environ["MY_AWS_DEFAULT_REGION"] == "us-east-2"


def test_to_env_os_system():
    # TODO: Implement
    pass


def test_to_env_popen():
    # TODO: Implement
    pass


def test_to_env_fork():
    # TODO: Implement
    pass


def test_to_env_execv():
    # TODO: Implement
    pass
